from __future__ import absolute_import

from pyros_interfaces_common.basenode import PyrosBase

from .mockinterface import MockInterface

# TODO : turn this into a main ? Since it is a process it might make it more explicit ??
# => can be run independently...


class PyrosMock(PyrosBase):
    """
    Mock Interface in pure python ( No ROS needed ).

    """
    #: Default configuration parameters.
    _default_config = {
        'TOPICS': [],
        'SERVICES': [],
        'PARAMS': [],
    }

    # TODO : we probably want to reuse the standard mock module here...
    def __init__(self, name=None, pyros_config=None, args=None, kwargs=None):
        name = name or 'pyros-mock'

        # No default config here for now

        # TODO : use context manager here to use another process just like any resource...
        super(PyrosMock, self).__init__(
            name,
            interface_class=MockInterface,
            args=args or (),
            kwargs=kwargs or {},
            default_config=self._default_config,  # we pass our default config here
        )

        # overriding default config with parameter provided
        # TMP currently disabling this to not depend on pyros. maybe it s not even needed ?
        # self.config_handler.configure(config)  # configuring with our package default
        if pyros_config:
            self.config_handler.configure(pyros_config)  # configuring with argument passed form user

        self._topic_msg = {}  # storage for the echo topic
        self._param_val = {}  # storage for the test param
        self._stop_event = None  # stop_event to signal the thread for soft shutdown
        self._spinner = None  # thread instance

    # These should match the design of PyrosClient and Protocol so we are consistent between pipe and python API
    def msg_build(self, name):
        # TODO : better mock : this can except if not string.
        msg = str()
        return msg

    # a simple echo topic
    def topic(self, name, msg_content=None):
        # TODO : use Mock interface topics directly
        msg = msg_content
        if msg_content is not None:
            self._topic_msg[name] = msg_content
            msg = None  # consuming the message
        else:
            msg = self._topic_msg.get(name)
        return msg
        
    def topics(self):
        """
        :return: the list of topics we interfaced with ( not the list of all available topics )
        """
        topics = self.interface.publishers.copy()
        topics.update(self.interface.subscribers)
        return topics

    # a simple publisher
    def subscriber(self, name, msg_content):
        # TODO : use Mock interface topics directly
        msg = msg_content
        self._topic_msg[name] = msg_content
        msg = None  # consuming the message
        return msg

    def subscribers(self):
        """
        :return: the list of topics we interfaced with ( not the list of all available topics )
        """
        return self.interface.subscribers

    def publishers(self):
        """
        :return: the list of topics we interfaced with ( not the list of all available topics )
        """
        return self.interface.publishers

    # a simple subscriber
    def publisher(self, name):
        # TODO : use Mock interface topics directly
        msg = self._topic_msg.get(name)
        return msg

    # a simple string echo service
    def service(self, name, rqst_content=None):
        # simulating a strict message typing backend
        #if not isinstance(rqst_content, str):
        #    raise Exception("Request Content Not Expected Type !")
        resp_content = rqst_content
        return resp_content

    def services(self):
        """
        :return: the list of services we interfaced with ( not the list of all available services )
        """
        return self.interface.services

    # a simple test param
    def param(self, name, value=None):
        val = value
        if value is not None:
            self._param_val[name] = value
            val = None  # consuming the message
        else:
            val = self._param_val.get(name)
        return val

    def params(self):
        """
        :return: the list of params we interfaced with ( not the list of all available params )
        """
        return self.interface.params

    def setup(self, publishers=None, subscribers=None, services=None, topics=None, params=None):
        """
        :param publishers:
        :param subscribers:
        :param services:
        :param topics: ONLY HERE for BW compat
        :param params:
        :return:
        """
        super(PyrosMock, self).setup(publishers=publishers, subscribers=subscribers, services=services, topics=topics, params=params)





# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import logging

"""
Hopefully this should endup in ros.__doc__
"""

# create logger
# TODO solve multiprocess logger problem(s)...
_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())



from __future__ import absolute_import

import logging
import six
import sys
import threading
import collections
import re
import abc
from functools import partial

# module wide to be pickleable
from contextlib import contextmanager

DiffTuple = collections.namedtuple("DiffTuple", " added removed ")

from .regex_tools import regex_match_sublist, regexes_match_sublist, find_first_regex_match


#TODO Entity Component System design for interface loop. cf https://pypi.python.org/pypi/esper (py3 + py2 in fork)
# Entities are transients (ex : for ROS : pubs, subs, svcs, params, and more can be added),
# Systems store logic about when/how a transient should be represented in the interface.
# GOALS : clarity, testability and flexibility
class TransientIfPool(object):
    # TODO : split available into TransientPool
    # TODO : have update and update_delta managed here in generic way...
    """

    The TransientIfPool class maintains :

    - Available transients collection:
    {
      transient1_name: transient1_class_instance,
      transient2_name: transient2_class_instance,
      transient3_name: transient3_class_instance,
    }

    - Interfaced transients collection:
    {
      transient1_name: transient_if1_class_instance,
      transient3_name: transient_if3_class_instance
    }

    update() checks (or get fed) the recent system state (or diff), and the interface store internally what transients are available.
    Also the interface creates/destroy the transients interface class instance when needed.

    Assumption : we only deal with absolute names here. The users should resolve them
    """
    __metaclass__ = abc.ABCMeta

    def update_transients(self, add_names, remove_names, *class_build_args, **class_build_kwargs):
        """
        Adds transients (service, topic, etc) named in add_names if they are not exposed in resolved_dict
        and removes transients named in remove_names if they are exposed in resolved_dict

        This method can be used to force exposing/withholding any transient, bypassing other mechanisms.
        No extra check is made here regarding regex or system status to try to guess if it should be added/removed.
        This is left ot the caller.

        :param add_names: the names of the transients to add
        :param remove_names: the names of the transients to remove
        :param class_build_args: the args to pass to the resolved transient constructor
        :param class_build_kwargs: the kwargs to pass to the resolved transient constructor
        :return: the list of transients exposed
        """
        # Important : no effect if names is empty list. only return empty. functional style.

        added = []
        removed = []
        for tst_name in [tst for tst in add_names if tst not in self.transients.keys()]:
            try:
                ttype = self.transient_type_resolver(tst_name)  # should return None if error - TODO : handle (and reraise) exception !!
                if ttype is not None:  # transient can be resolved
                    self.transients[tst_name] = self.TransientMaker(tst_name, ttype, *class_build_args, **class_build_kwargs)
                    added += [tst_name]
                    logging.info("[{name}] Interfacing with {desc} {transient}".format(name=__name__, desc=self.transients_desc, transient=tst_name))
                else:
                    logging.warning("[{name}] Type of {desc} {transient} unknown. Giving up trying to interface.".format(name=__name__, desc=self.transients_desc,
                                                                              transient=tst_name))
            except Exception as e:
                logging.warn("[{name}] Cannot interface with {desc} {transient} : {exc}".format(name=__name__, desc=self.transients_desc, transient=tst_name, exc=e))
                exc_info = sys.exc_info()
                six.reraise(exc_info[0], exc_info[1], exc_info[2])

        for tst_name in [tst for tst in remove_names if tst in self.transients.keys()]:
            if tst_name in self.transients:  # we make sure the transient is still exposed
                # because we might have modified resolved_dict after building the list to loop on
                logging.info("[{name}] Removing {desc} {transient}".format(name=__name__, desc=self.transients_desc, transient=tst_name))
                self.TransientCleaner(self.transients[tst_name])  # calling the cleanup function in case we need to do something
                self.transients.pop(tst_name, None)
                removed += [tst_name]

        return DiffTuple(added, removed)

    def expose_transients_regex(self, regexes, *class_build_args, **class_build_kwargs):
        """
        Exposes a list of transients regexes. resolved transients not matching the regexes will be removed.
        _expose_transients_regex -> _transient_change_detect -> _transient_change_diff -> _update_transients
        :param regexes: the list of regex to filter the transient to add.
               Note: regexes = [] remove all registered regexes.
        :param regex_set: the list of regex already existing
        :return: the list of transients exposed
        """
        # Important : no effect if names is empty list. only return empty diff. functional style.

        # look through the new service names received by reconfigure, and add
        # those services which are not in the existing service args
        for tst_regex in [r for r in regexes if not r in self.transients_args]:
            self.transients_args.add(tst_regex)
            logging.info('[{name}] Exposing {desc} regex : {regex}'.format(name=__name__, desc=self.transients_desc, regex=tst_regex))
            # TODO : check here for bugs & add test : what if we add multiple regexes ? wont we miss some add_names ?

        # look through the current service args and delete those values which
        # will not be valid when the args are replaced with the new ones. run on
        # a copy so that we will remove from the original without crashing
        for tst_regex in [r for r in self.transients_args if not r in regexes]:
            logging.info('[{name}] Withholding {desc} regex : {regex}'.format(name=__name__, desc=self.transients_desc, regex=tst_regex))
            self.transients_args.remove(tst_regex)

        return self.transient_change_detect(
            *class_build_args,
            **class_build_kwargs
         )

    # TODO: pass hte status as arguments instead of using self.get_transient_avialable
    def transient_change_detect(self, *class_build_args, **class_build_kwargs):
        """
        This should be called when we want to detect a change in the status of the system regarding the transient list
        This function also applies changes due to regex_set updates if needed
        _transient_change_detect -> _transient_change_diff -> _update_transients
        """

        transient_detected = set(self.get_transients_available())
        #TODO : unify that last_got_set with the *_available. they are essentially the same
        tst_gone = self.last_transients_detected - transient_detected

        # print("INTERFACING + {transient_detected}".format(**locals()))
        # print("INTERFACING - {tst_gone}".format(**locals()))

        dt = self.transient_change_diff(
            # we start interfacing with new matches,
            # but we also need to update old matches that match regex now
            transient_appeared=transient_detected,
            transient_gone=tst_gone,
            *class_build_args,
            **class_build_kwargs
        )

        self.last_transients_detected.update(transient_detected)
        if tst_gone:
            self.last_transients_detected.difference_update(tst_gone)

        return dt

    def transient_change_diff(self, transient_appeared, transient_gone, *class_build_args, **class_build_kwargs):
        """
        This should be called when we want to process a change in the status of the system (if we already have the diff)
        This function also applies changes due to regex_set updates if needed
        _transient_change_diff -> _update_transients
        """

        to_add = set(regexes_match_sublist(self.transients_args, transient_appeared))
        lost_matches = {n for n in self.transients if find_first_regex_match(n, self.transients_args) is None}
        to_remove = set(transient_gone) | lost_matches  # we stop interfacing with lost transient OR lost matches

        return self.update_transients(
            add_names=to_add,
            remove_names=to_remove,
            *class_build_args,
            **class_build_kwargs
        )

    # Abstract methods to override for services ("RPC / request / call type" communication channel)
    @abc.abstractmethod
    def get_transients_available(self):  # function returning all services available on the system
        return

    @abc.abstractmethod
    def transient_type_resolver(self, transient_name):  # function resolving the type of a service
        """
        :param service_name: the name of the service
        :return: returns None if the type cannot be found. Properly except in all other unexpected events.
        """
        return

    @abc.abstractmethod
    def TransientMaker(self, transient_name, service_type, *args, **kwargs):  # the service class implementation
        return

    @abc.abstractmethod
    def TransientCleaner(self, transient):  # the transient class implementation
        return

    def __init__(self, transients=None, transients_desc=None):
        """
        Initializes the interface instance, to expose services, topics, and params
        """

        # This is a local cache of the system state because we don't want to ask everytime.
        # TODO: find a way to make interface development easier by allowing developer to compare and worry only about local state representation versus interface
        # NOT about how to synchronize remote state with local state...
        self.available = dict()
        # This stays always in sync with the system (via interface update call)
        # but the interface instance can be created and recreated independently

        # Current transients exposed, i.e. those which are
        # active in the system.
        self.transients = {}
        self.transients_desc = transients_desc

        # Last requested transients to be exposed, received
        # from a setup request.
        self.transients_args = set()

        # to allow comparison with previously detected services / topics/ params
        # TODO : get rid of this
        self.last_transients_detected = set()

        # First update() will re-detect previously found names, but when trying to update, no duplicates should happen.
        # This should be taken care of in the abstract transient functional interface

        if transients is not None:
            tdt = self.expose_transients_regex(transients)
        else:
            tdt = DiffTuple([], [])  # no change in exposed services

    # TODO : implement this to abstract from lower logic
    # TODO : split this into a separate transient_pool
    # def reset_state(self, transients, transients_type):
    #     """
    #     Reset internal system state representation.
    #     expect lists in format similar to masterAPI. <= TODO : change this so data transform is done once only, outside of generic code
    #     :param transients:
    #     :param transients_data:
    #     :return:
    #     """
    #     self.available = dict()
    #     for s in services:  # We assume s[1] is never empty here
    #         st = next(ifilter(lambda lst: s[0] == lst[0], service_types), [])
    #         stp = ServiceTuple(name=s[0], type=st[1] if len(st) > 0 else None)
    #         self.available[stp.name] = stp
    #
    #     # We still need to return DiffTuples
    #     return services
    #
    # def compute_state(self, transients_added, transients_removed, service_types_dt):
    #     """
    #     This is called only if there is a cache proxy with a callback, and expects DiffTuple filled up with names or types
    #     :param services_dt:
    #     :param publishers_dt:
    #     :param subscribers_dt:
    #     :return:
    #     """
    #     for s in services_dt.added:
    #         st = next(ifilter(lambda lst: s[0] == lst[0], service_types_dt.added), [])
    #         stp = ServiceTuple(name=s[0], type=st[1] if len(st) > 0 else None)
    #         if stp.name in self.available:
    #             if self.available[stp.name].type is None and stp.type is not None:
    #                 self.available[stp.name].type = stp.type
    #         else:
    #             self.available[stp.name] = stp
    #
    #     for s in services_dt.removed:
    #         st = next(ifilter(lambda lst: s[0] == lst[0], service_types_dt.removed), [])
    #         stp = ServiceTuple(name=s[0], type=st[1] if len(st) > 0 else None)
    #         if stp.name in self.available:
    #             self.available.pop(stp.name, None)
    #
    #     # We still need to return DiffTuples
    #     return services_dt


    @contextmanager
    def mock(self, tst_name, tst_type):
        """
        Mock appearance/disappearance to force changes in interface , via setting local cache.
        :param tst_name:
        :param tst_type:
        :return:
        """
        print(" -> Mock {tst_name} appear".format(**locals()))
        # Service appears
        self.available[tst_name] = tst_type
        yield
        # Service disappear
        self.available.pop(tst_name)
        print(" -> Mock {tst_name} disappear".format(**locals()))

    # def update_on_diff(self, transients_dt):
    #
    #     # For added transients we need to check they match the regex, otherwise we skip it.
    #     # For removed transients we need to check that they have disappeared from system before removing, otherwise we skip it.
    #     # TODO : investigate shutdown behavior more in details
    #     # if shutting_down:
    #     #     # We force removing all interfaces by calling directly update services
    #     #     sdt = self.update_services(add_names=[], remove_names=[s for s in self.services])
    #     #     tdt = self.update_topics(add_names=[], remove_names=[t for t in self.topics])
    #     #     pdt = self.update_params(add_names=[], remove_names=[p for p in self.params])
    #     #else:
    #     tdt = self.update_services(add_names=regexes_match_sublist(self.transients_args, transients_dt.added),
    #                                remove_names=[s for s in transients_dt.removed if s not in self.get_transients_available()]
    #                                )
    #
    #     return DiffTuple(
    #         added=tdt.added,
    #         removed=tdt.removed,
    #     )


    # TODO : "wait_for_it" methods that waits for hte detection of a topic/service on the system
    # TODO : Should return a future so use can decide to wait on it or not
    # TODO : Maybe similar to a async_detect ( hooked up to the detected transient, not the exposed ones )
    # TODO : Exposed interface is for direct control flow => async not really needed
    # TODO : Detect/Update interface is inversed control flow ( from update loop ) => Needed





# -*- coding: utf-8 -*-
from __future__ import absolute_import

#
# Simulating an external environment
#

from contextlib import contextmanager
import multiprocessing
from collections import namedtuple


TopicType = namedtuple("TopicType", "msgtype")

statusecho_topic = TopicType("StatusMsg")


# This manager process provides a mockinterface implementation while allowing multiprocess access to it
# Especially useful for test client process.
mock_manager = multiprocessing.Manager()

# These list what is available on the Mock implementation.
# They are accessible directly for tests who want to simulate multiprocess communication framework changes.
services_available_remote = mock_manager.list()  # Service is unique
services_available_type_remote = mock_manager.dict()

# Same type for pub and subs
topics_available_remote = mock_manager.dict()  # instance Counter ( Potentially N Pubs + M Subs )
topics_available_type_remote = mock_manager.dict()

params_available_remote = mock_manager.list()  # Param is unique
params_available_type_remote = mock_manager.dict()


@contextmanager
def mock_service_remote(svc_name, svc_type):
    print(" -> Mock Service {svc_name} appear".format(**locals()))
    services_available_remote.append(svc_name)  # Service appears
    services_available_type_remote[svc_name] = svc_type
    yield
    services_available_remote.remove(svc_name)
    services_available_type_remote.pop(svc_name)
    print(" -> Mock Service {svc_name} disappear".format(**locals()))


@contextmanager
def mock_publisher_remote(topic_name, topic_type):
    print(" -> Mock Topic {topic_name} appear".format(**locals()))
    topics_available_remote[topic_name] = topics_available_remote.get(topic_name, 0) + 1  # Publisher appears
    if topic_name in topics_available_type_remote:
        assert(topics_available_type_remote[topic_name] == topic_type)
    else:
        topics_available_type_remote[topic_name] = topic_type
    yield
    topics_available_remote[topic_name] = topics_available_remote.get(topic_name, 0) - 1  # Publisher disappears
    if topics_available_remote.get(topic_name, 0) == 0:
        topics_available_remote.pop(topic_name)
        topics_available_type_remote.pop(topic_name)
    print(" -> Mock Topic {topic_name} disappear".format(**locals()))

@contextmanager
def mock_subscriber_remote(topic_name, topic_type):
    print(" -> Mock Topic {topic_name} appear".format(**locals()))
    topics_available_remote[topic_name] = topics_available_remote.get(topic_name, 0) + 1  # Subscriber appears
    if topic_name in topics_available_type_remote:
        assert(topics_available_type_remote[topic_name] == topic_type)
    else:
        topics_available_type_remote[topic_name] = topic_type
    yield
    topics_available_remote[topic_name] = topics_available_remote.get(topic_name, 0) - 1  # Subscriber disappears
    if topics_available_remote.get(topic_name, 0) == 0:
        topics_available_remote.pop(topic_name)
        topics_available_type_remote.pop(topic_name)
    print(" -> Mock Topic {topic_name} disappear".format(**locals()))

@contextmanager
def mock_param_remote(param_name, param_type):
    print(" -> Mock Param {param_name} appear".format(**locals()))
    params_available_remote.append(param_name)  # Param appears
    params_available_type_remote[param_name] = param_type
    yield
    params_available_remote.remove(param_name)  # Param disappears
    params_available_type_remote.pop(param_name)
    print(" -> Mock Param {param_name} disappear".format(**locals()))


from .mockpublisher import MockPublisher
from .mocksubscriber import MockSubscriber
from .mockservice import MockService
from .mockparam import MockParam


# If we prefer a in process class instance instead of a remote process
# BUT remote process is used by Pyros_Mock
# TODO : choose ONE !
# TODO : find a simpler way to insure isolation for Mock...
# TODO : investigate how to mix / reuse standard python multprocessing lib
# Check SyncManager http://stackoverflow.com/questions/19568232/how-to-iterate-over-a-dict-proxy-in-python

class MockSystem(object):
    services_available_remote = []  # Service is unique
    services_available_type_remote = {}

    # Same type for pub and subs
    topics_available_remote = {}  # instance Counter ( Potentially N Pubs + M Subs )
    topics_available_type_remote = {}

    params_available_remote = []  # Param is unique
    params_available_type_remote = {}

    # Mock Implementation
    # Intra-process inter-thread communication channel : a simple global variable
    _msg_subs = dict()

    @contextmanager
    def mock_service_remote(self, svc_name, svc_type):
        print(" -> Mock Service {svc_name} appear".format(**locals()))
        self.services_available_remote.append(svc_name)  # Service appears
        self.services_available_type_remote[svc_name] = svc_type
        yield
        self.services_available_remote.remove(svc_name)
        self.services_available_type_remote.pop(svc_name)
        print(" -> Mock Service {svc_name} disappear".format(**locals()))

    @contextmanager
    def mock_publisher_remote(self, topic_name, topic_type):
        print(" -> Mock Topic {topic_name} appear".format(**locals()))
        self.topics_available_remote[topic_name] = self.topics_available_remote.get(topic_name, 0) + 1  # Publisher appears
        if topic_name in self.topics_available_type_remote:
            assert (self.topics_available_type_remote[topic_name] == topic_type)
        else:
            self.topics_available_type_remote[topic_name] = topic_type
        yield
        self.topics_available_remote[topic_name] = self.topics_available_remote.get(topic_name, 0) - 1  # Publisher disappears
        if self.topics_available_remote.get(topic_name, 0) == 0:
            self.topics_available_remote.pop(topic_name)
            topics_available_type_remote.pop(topic_name)
        print(" -> Mock Topic {topic_name} disappear".format(**locals()))

    @contextmanager
    def mock_subscriber_remote(self, topic_name, topic_type):
        print(" -> Mock Topic {topic_name} appear".format(**locals()))
        self.topics_available_remote.append(topic_name)  # Service appears
        self.topics_available_type_remote[topic_name] = topic_type
        yield
        self.topics_available_remote.remove(topic_name)  # Service disappears
        self.topics_available_type_remote.pop(topic_name)
        print(" -> Mock Topic {topic_name} disappear".format(**locals()))

    @contextmanager
    def mock_param_remote(self, param_name, param_type):
        print(" -> Mock Param {param_name} appear".format(**locals()))
        self.params_available_remote.append(param_name)  # Param appears
        self.params_available_type_remote[param_name] = param_type
        yield
        self.params_available_remote.remove(param_name)  # Param disappears
        self.params_available_type_remote.pop(param_name)
        print(" -> Mock Param {param_name} disappear".format(**locals()))

    # TODO : define a better API (more similar to rospy / ROS interfaces...)
    # TODO : link with .api module somehow...
    def create_publisher(self, topic_name, topic_type):
        pub = MockPublisher(topic_name=topic_name, topic_type=topic_type)
        pub.system = self  # to enable topic comm
        return pub

    def create_subscriber(self, topic_name, topic_type):
        sub = MockSubscriber(topic_name=topic_name, topic_type=topic_type)
        self._msg_subs[sub.name] = self._msg_subs.get(sub.name, []) + [sub]
        return sub

    def create_service(self, service_name, service_type):
        svc = MockService(service_name=service_name, service_type=service_type)
        return svc

    def create_parameter(self, param_name, param_type):
        prm = MockParam(param_name=param_name, param_type=param_type)
        return prm

    def transfer(self, msg, topic_name):
        for sub in self._msg_subs.get(topic_name, []):
            sub.topic_callback(msg)
        return True


__all__ = [

    'services_available_remote',
    'services_available_type_remote',
    'mock_service_remote',

    'topics_available_remote',
    'topics_available_type_remote',
    'mock_topic_remote'

    'params_available_remote',
    'params_available_type_remote',
    'mock_param_remote',
]



from __future__ import absolute_import

from pyros_interfaces_common.transient_if_pool import TransientIfPool
from .mockservice import MockService
from .mocksystem import (
    services_available_remote, services_available_type_remote,
)


class MockServicePool(TransientIfPool):

    """
    MockInterface.
    """
    def __init__(self, services=None):
        # This base constructor assumes the system to interface with is already available ( can do a get_svc_available() )
        super(MockServicePool, self).__init__(services)

    # mock functions that simulate/mock similar interface than what is found on multiprocess framework supported
    # We should try our best to go for the lowest common denominator here
    # SERVICES
    def get_transients_available(self):  # function returning all services available on the system
        return self.available

    def transient_type_resolver(self, service_name):  # function resolving the type of a service
        svc = self.available.get(service_name)
        return svc  # None is returned if not found

    def TransientMaker(self, service_name, service_type, *args, **kwargs):  # the service class implementation
        return MockService(service_name, service_type, *args, **kwargs)

    def TransientCleaner(self, service):  # the service class cleanup implementation
        return service.cleanup()

    def update(self):
        for s in services_available_remote:
            self.available[s] = services_available_type_remote.get(s)

        dt = self.transient_change_detect()

        return dt





from __future__ import absolute_import

import abc
import contextlib
import importlib

import pyzmp
from pyros_config import ConfigHandler

from pyros_common.exceptions import PyrosException


# TODO : cleaner design by using pyzmp.Node as delegate to make all interface explicit here...
class PyrosBase(pyzmp.Node):
    """
    Base Interface in pure python ( No ROS needed ).
    Encapsulate another process for isolated execution
    Also handles a package bound configuration (since the child process might want to reload it in its own memory space)
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self,
                 name=None,
                 interface_class=None,
                 context_manager=None,
                 args=None,
                 kwargs=None,
                 instance_path=None,
                 instance_relative_config=True,
                 root_path=None,
                 default_config=None):
        """
        :param name: name of the node
        :param interface_class: the class of the interface to instantiate (in child process)
                OR a tuple (module_name, class_name), useful if the module should be imported only in child process
                OR a tuple (package_name, module_name, class_name), usseful if the module name is relative
                The interface class should be a child of BaseInterface.
        :param context_manager: a context manager to enter when starting, and exit when the node stops
        :param args: the args to pass to the setup function (that instantiate the interface class)
        :param kwargs: the kwargs to pass to the setup function (that instantiate the interface class)
        :param instance_path: the path to the running instance of the running app. configurations and resources will be expected there.
                By default it will automatically use a suitable 'instance/' folder.
                The configuration strategy follows flask configuration strategy.
                More information there : http://flask.pocoo.org/docs/0.10/config/#instance-folders
        :param instance_relative_config: whether the configuration is relative to the instance_path (if not it will be relative to the root_path)
        :param root_path: the root_path. by default the package location.
        :param default_config: a dict containing the values to use as default configuration values
        :return:
        """
        super(PyrosBase, self).__init__(name or 'pyros', context_manager=context_manager, args=args or (), kwargs=kwargs or {})

        self.last_update = 0
        self.update_interval = 1  # seconds to wait between each update

        # we delegate config related behavior (including defaults)
        self.config_handler = ConfigHandler(
            name or 'pyros',
            instance_path=instance_path,
            instance_relative_config=instance_relative_config,
            root_path=root_path,
            default_config=default_config,
        )

        self.provides(self.msg_build)
        # BWCOMPAT
        self.provides(self.topic)
        self.provides(self.topics)
        #
        self.provides(self.publisher)
        self.provides(self.publishers)
        self.provides(self.subscriber)
        self.provides(self.subscribers)
        self.provides(self.service)
        self.provides(self.services)
        self.provides(self.param)
        self.provides(self.params)
        self.provides(self.setup)


        if not isinstance(interface_class, tuple) and not (
                # TODO : we should pre check all the used members are present...
                hasattr(interface_class, 'services')
                # TODO : etc.
        ):
            raise PyrosException("The interface passed to the Node is not a tuple or is missing some members to be used as an interface : {interface_class}".format(**locals()))
        else:
            self.interface_class = interface_class
        self.interface = None  # delayed interface creation
        # interface should be created in child process only to maintain isolation.
        # BUT in a context to have everything setup after returning from start() call.

    #
    # Delegating configuration management
    #

    @property
    def import_name(self):
        return self.config_handler.import_name

    @property
    def config(self):
        return self.config_handler.config

    @property
    def instance_path(self):
        return self.config_handler.instance_path

    @property
    def instance_relative_config(self):
        return self.config_handler.instance_relative_config

    @property
    def root_path(self):
        return self.config_handler.root_path

    # TODO : review API : configure and setup are similar, but on different processes, at different times...
    def configure(self, config=None):
        # We default to using a config file named after the import_name:
        config = config or self.name + '.cfg'
        self.config_handler.configure(config)
        return self

    #
    # These should match the design of PyrosClient and Protocol so we are consistent between pipe and python API
    #
    @abc.abstractmethod
    def msg_build(self, name):
        return

    # BWCOMPAT
    @abc.abstractmethod
    def topic(self, name, msg_content=None):
        return

    @abc.abstractmethod
    def topics(self):
        return

    @abc.abstractmethod
    # CAREFUL : accessing a publisher means we want to receive a message from it
    def publisher(self, name):
        return

    @abc.abstractmethod
    def publishers(self):
        return

    @abc.abstractmethod
    # CAREFUL : accessing a subscriber means we want to send message to it
    def subscriber(self, name, msg_content):
        return

    @abc.abstractmethod
    def subscribers(self):
        return

    # a simple string echo service
    @abc.abstractmethod
    def service(self, name, rqst_content=None):
        return

    @abc.abstractmethod
    def services(self):
        return

    @abc.abstractmethod
    def param(self, name, value=None):
        return

    @abc.abstractmethod
    def params(self):
        return

    def start(self, timeout=None):
        """
        Startup of the node.
        :param join: optionally wait for the process to end (default : True)
        :return: None
        """

        assert super(PyrosBase, self).start(timeout=timeout)
        # Because we currently use this to setup connection
        return self.name

    @abc.abstractmethod
    def setup(self, *args, **kwargs):
        """
        Dynamically reset the interface to expose the services / topics / params whose names are passed as args
        The interface class can be specified with a module to be dynamically imported
        :param publishers:
        :param subscribers:
        :param services:
        :param topics: BW COMPAT ONLY !
        :param params:
        :return:
        """

        #BWCOMPAT
        if kwargs.get('topics'):
            if kwargs.get('publishers'):
                kwargs['publishers'] = kwargs.get('publishers', []) + kwargs.get('topics', [])
            else:
                kwargs['publishers'] = kwargs.get('topics', [])
            if kwargs.get('subscribers'):
                kwargs['subscribers'] = kwargs.get('subscribers', []) + kwargs.get('topics', [])
            else:
                kwargs['subscribers'] = kwargs.get('topics', [])
        kwargs.pop('topics')

        # We can now import RosInterface and setup will be done ( we re in another process ).
        # TODO : context to make it cleaner (maybe use zmp.Node context ?)
        if isinstance(self.interface_class, tuple):
            m = None
            class_name = self.interface_class[-1]  # last element is always the class_name
            if len(self.interface_class) >= 3:
                # load the relative module, will raise ImportError if module cannot be loaded
                m = importlib.import_module(self.interface_class[1], self.interface_class[0])
            elif len(self.interface_class) == 2:
                # load the relative module, will raise ImportError if module cannot be loaded
                m = importlib.import_module(self.interface_class[0])
            # get the class, will raise AttributeError if class cannot be found
            self.interface_class = getattr(m, class_name)

        if not (
                # TODO : we should pre check all the used members are present...
                hasattr(self.interface_class, 'services')
                # TODO : etc.
        ):
            raise PyrosException("The interface class is missing some members to be used as an interface. Aborting Setup. {interface_class}".format(**locals()))

        self.interface = self.interface_class(*args, **kwargs)
        return self.interface

    @contextlib.contextmanager
    def child_context(self, *args, **kwargs):
        """
        Context setup first in child process, before returning from start() call in parent.
        Result is passed in as argument of update
        :return:
        """

        # Now we can extract config values
        expected_args = {
            'services': [],
            'topics': [],  # bwcompat
            'subscribers': [],
            'publishers': [],
            'params': [],
            # TODO : all of them !
        }

        ifargs = {
            arg: self.config_handler.config.get(arg.upper(), default) for arg, default in expected_args.items()
        }

        # overriding with kwargs
        ifargs.update(kwargs)

        # storing passed args in config in case of reset

        # calling setup on child context enter call
        if self.interface is None:
            #for BW compat
            # TODO : change API to use the child_context from pyzmp coprocess
            self.setup(*args, **ifargs)

        with super(PyrosBase, self).child_context(*args, **kwargs) as cctxt:
            yield cctxt

    def shutdown(self, join=True, timeout=None):
        """
        Clean shutdown of the node.
        :param join: optionally wait for the process to end (default : True)
        :return: last exitcode from update method
        """
        if self.interface is not None:
            self.interface.stop()

        return super(PyrosBase, self).shutdown(join, timeout=timeout)

    def update(self, timedelta, *args, **kwargs):
        """
        Update function to call from a looping thread.
        Note : the interface is lazily constructed here
        :param timedelta: the time past since the last update call
        """

        # TODO move time management somewhere else...
        self.last_update += timedelta
        # if shutdown we want to bypass the update_interval check
        # if shutting_down:
        #     self.interface.update(shutting_down=shutting_down)
        #     return 0  # return 0 as success since we arrived here without exception
        # else
        if self.last_update > self.update_interval:
            self.last_update = 0
            self.interface.update()

        # No return here means we need to keep looping





from __future__ import absolute_import

import os



# All Pyros Exception must be pickleable and have a message property
class PyrosException(Exception):
    """Base exception for errors raised by Pyros
       Initialize it with members that can be pickled when moving the exception around
       Redefine the message property to get a meaningful message:

       >>>  # TODO doctest
    """

    def __init__(self, *args):
        """
        Overriding Exception constructor to handle both old and new Exception format...
        :param args:
        :return:
        """
        super(PyrosException, self).__init__(*args)

    # redefining __str__ to get custom message to appear on raise
    def __str__(self):
        return self.message

    @property
    def message(self):
        """
        message as part of core Python Exception has been deprecated.
        We recycle it here for our own use.

        It is not a member, so it will not be serialized and transferred between processes
        Instead it will be reconstructed from members ( we have shared the code for it )
        :return:
        """
        return "An Exception {0} was thrown in Pyros process {1}".format(repr(self), os.getpid())




from __future__ import absolute_import, print_function

import sys

if sys.version_info < (3, 0):
    import __builtin__ as builtins
else:
    import builtins

from collections import namedtuple

# TODO : use basemsg and develop a generic way of converting python type to custom type


if sys.version_info < (3, 0):
    # type map specifying conversion
    # basic types should be python builtin types
    # if not in there : assumed to be a custom type, convertible to a dict.
    type_map = {
       "bool":    ["bool"],
       "int":     ["int", "float"],  # can convert int to float
       "float":   ["float"],
       "str":     ["str", "unicode"],  # convert str to unicode to follow python 3 default behavior
       "unicode": ["unicode"],
       "list":    ["list"],
       "tuple":   ["tuple", "list"],  # can convert tuple to list
    }

    # Starting with whatever ROS needs, but we could extend this
    primitive_types = [bool, int, long, float, str, unicode]
    composed_types = [list, tuple]
else:  # python3
    # type map specifying conversion
    # basic types should be python builtin types
    # if not in there : assumed to be a custom type, convertible to a dict.
    type_map = {
        "bool": ["bool"],
        "int": ["int", "float"],  # can convert int to float
        "float": ["float"],
        "bytes": ["bytes", "str"],  # bytes can be converted to str
        "str": ["str"],
        "list": ["list"],
        "tuple": ["tuple", "list"],  # can convert tuple to list
    }

    primitive_types = [bool, int, float, bytes, str]
    composed_types = [list, tuple]

# defining Mock message types using namedtuple to keep things small
StatusMsg = namedtuple("StatusMsg", "error code message")

# defining Exceptions
class NonexistentFieldException(Exception):
    def __init__(self, oridict, basetype, message):
        Exception.__init__(self, "Trying to convert {0!s} to Message type {1!s} triggered {2!s}".format(oridict, basetype, message))


class FieldTypeMismatchException(Exception):
    def __init__(self, roottype, fields, expected_type, found_type):
        if roottype == expected_type:
            Exception.__init__(self, "Expected a Python object for type {0!s} but received a {1!s}".format(roottype, found_type))
        else:
            Exception.__init__(self, "{0!s} message requires a {1!s} for field {2!s}, but got a {3!s}".format(roottype, expected_type, '.'.join(fields), found_type))


def extract_values(inst):
    """
    :param inst: the instance
    :return: python values extracted from the instance
    """
    # inst should already be python
    return inst


def populate_instance(msg, inst):
    """
    :param msg: contains the values to use to populate inst.
    :param inst: message class instance to populate.
    :return: an instance of the provided message class, with its fields populated according to the values in msg
    """
    return _to_inst(msg, type(inst).__name__, type(inst).__name__, inst)


def _to_inst(msg, ptype, roottype, inst=None, stack=None):

    # Check to see whether this is a primitive type
    if stack is None:
        stack = []
    if ptype in builtins.__dict__ and (
        builtins.__dict__[ptype] in primitive_types or
        builtins.__dict__[ptype] in composed_types
    ):
        # Typecheck the msg
        msgtype = type(msg)

        if msgtype in primitive_types and ptype in type_map[msgtype.__name__]:
            return builtins.__dict__[ptype](msg)
        elif msgtype in composed_types and ptype in type_map[msgtype.__name__]:
            # Call to _to_inst for every element of the list/tuple
            def recurse_iter(msg):
                for e in msg:  # we do this with yield to get an iteratable and build the tuple/list at once
                    yield _to_inst(e, type(e).__name__, roottype, None, stack)

            return builtins.__dict__[ptype](recurse_iter(msg))

        raise FieldTypeMismatchException(roottype, stack, ptype, msgtype)

    # Otherwise, the type has to be a custom type, so msg must be a dict
    if type(msg) is not dict:
        raise FieldTypeMismatchException(roottype, stack, ptype, type(msg))

    # and ptype should be able to build with same fields.

    # modifying dict with dict comprehension
    instmsg = dict((k, _to_inst(v, type(v).__name__, roottype, None, stack)) for k, v in msg.items())

    # using ** to get dict content as named args for the namedtuple
    try:
        inst = globals()[ptype](**instmsg)
    except TypeError as e:
        raise NonexistentFieldException(msg, ptype, e)
    return inst



from __future__ import absolute_import

from .mockservicepool import MockServicePool
from .mocksubscriberpool import MockSubscriberPool
from .mockpublisherpool import MockPublisherPool
from .mockparampool import MockParamPool


# TODO : merge this with mock interface... Goal is to split Mock and ROS interfaces...

# TODO Entity Component System design for interface loop. cf https://pypi.python.org/pypi/esper (py3 + py2 in fork)
# Entities are transients (ex : for ROS : pubs, subs, svcs, params, and more can be added),
# Systems store logic about when/how a transient should be represented in the interface.
# GOALS : clarity, testability and flexibility
class BaseInterface(object):

    """
    BaseInterface.
    Assumption : we only deal with absolute names here. The users should resolve them
    """

    # Abstract methods to override for topics ("pub/sub type" communication channel)
    def get_publisher_list(self):  # function returning all topics available on the system
        return self.publishers_if_pool.get_transients_available()

    def publisher_type_resolver(self, topic_name):  # function resolving the type of a topic
        """
        :param topic_name: the name of the topic
        :return: returns None if the topic cannot be found. Properly except in all other unexpected events.
        """
        return self.publishers_if_pool.transient_type_resolver(topic_name)

    def PublisherMaker(self, topic_name, topic_type, *args, **kwargs):  # the topic class implementation
        return self.publishers_if_pool.TransientMaker(topic_name, topic_type, *args, **kwargs)

    def PublisherCleaner(self, topic):  # the topic class implementation
        return self.publishers_if_pool.TransientCleaner(topic)

    # Abstract methods to override for topics ("pub/sub type" communication channel)
    def get_subscriber_list(self):  # function returning all topics available on the system
        return self.subscribers_if_pool.get_transients_available()

    def subscriber_type_resolver(self, topic_name):  # function resolving the type of a topic
        """
        :param topic_name: the name of the topic
        :return: returns None if the topic cannot be found. Properly except in all other unexpected events.
        """
        return self.subscribers_if_pool.transient_type_resolver(topic_name)

    def SubscriberMaker(self, topic_name, topic_type, *args, **kwargs):  # the topic class implementation
        return self.subscribers_if_pool.TransientMaker(topic_name, topic_type, *args, **kwargs)

    def SubscriberCleaner(self, topic):  # the topic class implementation
        return self.subscribers_if_pool.TransientCleaner(topic)

    def get_svc_list(self):  # function returning all services available on the system
        return self.services_if_pool.get_transients_available()

    def service_type_resolver(self, service_name):  # function resolving the type of a service
        """
        :param service_name: the name of the service
        :return: returns None if the type cannot be found. Properly except in all other unexpected events.
        """
        return self.services_if_pool.transient_type_resolver(service_name)

    def ServiceMaker(self, service_name, service_type, *args, **kwargs):  # the service class implementation
        return self.services_if_pool.TransientMaker(service_name, service_type, *args, **kwargs)

    def ServiceCleaner(self, service):  # the service class implementation
        return self.services_if_pool.TransientCleaner(service)

    # Abstract methods to override for topics ("pub/sub type" communication channel)
    def get_topic_list(self):  # function returning all topics available on the system
        return self.topics_pool.get_transients_available()

    def topic_type_resolver(self, topic_name):  # function resolving the type of a topic
        """
        :param topic_name: the name of the topic
        :return: returns None if the topic cannot be found. Properly except in all other unexpected events.
        """
        return self.topics_pool.transient_type_resolver(topic_name)

    def TopicMaker(self, topic_name, topic_type, *args, **kwargs):  # the topic class implementation
        return self.topics_pool.TransientMaker(topic_name, topic_type, *args, **kwargs)

    def TopicCleaner(self, topic):  # the topic class implementation
        return self.topics_pool.TransientCleaner(topic)

    # Abstract methods to override for params ( "global" get/set objects )
    def get_param_list(self):  # function returning all topics available on the system
        return self.params_if_pool.get_transients_available()

    def param_type_resolver(self, param_name):  # function resolving the type of a topic
        return self.params_if_pool.transient_type_resolver(param_name)

    def ParamMaker(self, param_name, param_type, *args, **kwargs):  # the param class implementation
        return self.params_if_pool.TransientMaker(param_name, param_type, *args, **kwargs)

    def ParamCleaner(self, param):  # the param class implementation
        return self.params_if_pool.TransientCleaner(param)

    def __init__(self, publishers_if_pool, subscribers_if_pool, services_if_pool, params_if_pool):
        """
        Initializes the interface instance, to expose services, topics, and params
        """
        # Current transients exposed, i.e. those which are
        # active in the system.

        self.params_if_pool = params_if_pool
        self.services_if_pool = services_if_pool
        self.subscribers_if_pool = subscribers_if_pool
        self.publishers_if_pool = publishers_if_pool

    # REQUESTED
    @property
    def publishers_args(self):
        return self.publishers_if_pool.transients_args

    @property
    def subscribers_args(self):
        return self.subscribers_if_pool.transients_args

    @property
    def services_args(self):
        return self.services_if_pool.transients_args

    @property
    def params_args(self):
        return self.params_if_pool.transients_args

    # AVAILABLE
    @property
    def publishers_available(self):
        return self.publishers_if_pool.available

    @property
    def subscribers_available(self):
        return self.subscribers_if_pool.available

    @property
    def services_available(self):
        return self.services_if_pool.available

    @property
    def params_available(self):
        return self.params_if_pool.available

    # INTERFACED
    @property
    def publishers(self):
        return self.publishers_if_pool.transients

    @property
    def subscribers(self):
        return self.subscribers_if_pool.transients

    @property
    def services(self):
        return self.services_if_pool.transients

    @property
    def params(self):
        return self.params_if_pool.transients

    # EXPOSE
    def expose_publishers(self, tpc_regex):
        return self.publishers_if_pool.expose_transients_regex(tpc_regex)

    def expose_subscribers(self, tpc_regex):
        return self.subscribers_if_pool.expose_transients_regex(tpc_regex)

    def expose_services(self, svc_regex):
        return self.services_if_pool.expose_transients_regex(svc_regex)

    def expose_params(self, prm_regex):
        return self.params_if_pool.expose_transients_regex(prm_regex)

    #CHANGE DIFF
    def publishers_change_diff(self, appeared, gone):
        return self.publishers_if_pool.transient_change_diff(appeared, gone)

    def subscribers_change_diff(self, appeared, gone):
        return self.subscribers_if_pool.transient_change_diff(appeared, gone)

    def services_change_diff(self, appeared, gone):
        return self.services_if_pool.transient_change_diff(appeared, gone)

    def params_change_diff(self, appeared, gone):
        return self.params_if_pool.transient_change_diff(appeared, gone)

    #CHANGE DETECT
    def publishers_change_detect(self):
        return self.publishers_if_pool.transient_change_detect()

    def subscribers_change_detect(self):
        return self.subscribers_if_pool.transient_change_detect()

    def services_change_detect(self):
        return self.services_if_pool.transient_change_detect()

    def param_change_detect(self):
        return self.params_if_pool.transient_change_detect()

    #UPDATE
    def update_publishers(self, add_names, remove_names):
        return self.publishers_if_pool.update_transients(add_names, remove_names)

    def update_subscribers(self, add_names, remove_names):
        return self.subscribers_if_pool.update_transients(add_names, remove_names)

    def update_services(self, add_names, remove_names):
        return self.services_if_pool.update_transients(add_names, remove_names)

    def update_params(self, add_names, remove_names):
        return self.params_if_pool.update_transients(add_names, remove_names)

    # TODO : "wait_for_it" methods that waits for hte detection of a topic/service on the system
    # TODO : Should return a future so use can decide to wait on it or not
    # TODO : Maybe similar to a async_detect ( hooked up to the detected transient, not the exposed ones )
    # TODO : Exposed interface is for direct control flow => async not really needed
    # TODO : Detect/Update interface is inversed control flow ( from update loop ) => Needed



class MockInterface(BaseInterface):

    """
    MockInterface. A simple mock following ROS design.
    Userful to test the multiprocess parts of pyros without worrying about real connexion with ROS.
    """
    def __init__(self, publishers=None, subscribers=None, services=None, params=None):

        publishers_pool = MockPublisherPool(publishers)
        subscribers_pool = MockSubscriberPool(subscribers)
        services_pool = MockServicePool(services)
        params_pool = MockParamPool(params)

        super(MockInterface, self).__init__(publishers_pool, subscribers_pool, services_pool, params_pool)

    #
    # # REQUESTED
    # @property
    # def services_args(self):
    #     return self.service_pool.transients_args
    #
    # @property
    # def topics_args(self):
    #     return self.topic_pool.transients_args
    #
    # @property
    # def params_args(self):
    #     return self.param_pool.transients_args
    #
    # # AVAILABLE
    # @property
    # def services_available(self):
    #     return self.service_pool.available
    #
    # @property
    # def topics_available(self):
    #     return self.topic_pool.available
    #
    # @property
    # def params_available(self):
    #     return self.param_pool.available
    #
    # # INTERFACED
    # @property
    # def services(self):
    #     return self.service_pool.transients
    #
    # @property
    # def topics(self):
    #     return self.topic_pool.transients
    #
    # @property
    # def params(self):
    #     return self.param_pool.transients
    #
    # # EXPOSE
    # def expose_services(self, svc_regex):
    #     return self.service_pool.expose_transients_regex(svc_regex)
    #
    # def expose_topics(self, tpc_regex):
    #     return self.topic_pool.expose_transients_regex(tpc_regex)
    #
    # def expose_params(self, prm_regex):
    #     return self.param_pool.expose_transients_regex(prm_regex)
    #
    #MOCK
    def mock_publisher(self, tpc_name, tpc_type):
        return self.publishers_if_pool.mock(tpc_name, tpc_type)

    def mock_subscriber(self, tpc_name, tpc_type):
        return self.subscribers_if_pool.mock(tpc_name, tpc_type)

    def mock_service(self, svc_name, svc_type):
        return self.services_if_pool.mock(svc_name, svc_type)

    def mock_param(self, prm_name, prm_type):
        return self.params_if_pool.mock(prm_name, prm_type)
    #
    # #CHANGE DIFF
    # def services_change_diff(self, appeared, gone):
    #     return self.service_pool.transient_change_diff(appeared, gone)
    #
    # def topics_change_diff(self, appeared, gone):
    #     return self.topic_pool.transient_change_diff(appeared, gone)
    #
    # def params_change_diff(self, appeared, gone):
    #     return self.param_pool.transient_change_diff(appeared, gone)
    #
    # #CHAGNE DETECT
    # def services_change_detect(self):
    #     return self.service_pool.transient_change_detect()
    #
    # def topics_change_detect(self):
    #     return self.topic_pool.transient_change_detect()
    #
    # def param_change_detect(self):
    #     return self.param_pool.transient_change_detect()
    #
    # #UPDATE
    # def update_services(self, add_names, remove_names):
    #     return self.service_pool.update_transients(add_names, remove_names)
    #
    # def update_topics(self, add_names, remove_names):
    #     return self.topic_pool.update_transients(add_names, remove_names)
    #
    # def update_params(self, add_names, remove_names):
    #     return self.param_pool.update_transients(add_names, remove_names)

    def update(self):
        self.publishers_if_pool.update()
        self.subscribers_if_pool.update()
        self.services_if_pool.update()
        self.params_if_pool.update()



from __future__ import absolute_import, print_function

from pyros_interfaces_common.transient_if import TransientIf


class MockParam(TransientIf):
    """
    MockParam is the class handling mock behavior for Param
    """
    def __init__(self, param_name, param_type):

        # getting the fullname to make sure we start with /
        param_name = param_name if param_name.startswith('/') else '/' + param_name

        param_type = param_type()  # or None ??

        super(MockParam, self).__init__(param_name, param_type)

    def cleanup(self):
        pass

    def setval(self, val):
        self.value = val
        return True

    def getval(self):
        return self.value



from __future__ import absolute_import
from __future__ import print_function


class TransientIf(object):
    """
    TransientIf is a base class handling conversion from python API to Multiprocess transients.
    This transient can be a service, a param, a publisher, etc.
    This class aims at interfacing this transient with python in a common way for all multiprocess implementations.
    It will be manipulated from TransientPool, to track when this interface should be created or destroyed.
    """

    def __init__(self, transient_name, transient_type=None):
        """
        Building a transient from external data
        :param transient_name:
        :param transient_data:
        """

        self.name = transient_name
        self.type = transient_type

    def cleanup(self):
        """
        cleanup function for a transient interface
        :return:
        """
        return




from __future__ import absolute_import

import logging
from collections import namedtuple
"""
Pure python protocol ( only for pipe communication between process )
Designed to have always atomic sensible data on the pipe.
TODO : double check if this is still needed. ZMP will encapsulate everything needed...
"""

MsgBuild = namedtuple('MsgBuild', 'name msg_content')
# rqst : msg_content = None
# resp : msg_content = <msg_asdict>

Topic = namedtuple('Topic', 'name msg_content')
# Inject :
# rqst : msg_content = <msg_asdict>
# resp : msg_content = None
# Extract :
# rqst : msg_content = None
# resp : msg_content = <msg_asdict>

TopicList = namedtuple('TopicList', 'name_dict')
TopicInfo = namedtuple('TopicInfo', 'name fullname msgtype allow_sub allow_pub rostype rostype_name')

Service = namedtuple('Service', 'name rqst_content resp_content')
# rqst : rqst_content = <msg_asdict_1>, resp_content = None
# resp : rqst_content = <msg_asdict_1>, resp_content = <msg_asdict>

ServiceList = namedtuple('ServiceList', 'name_dict')
ServiceInfo = namedtuple('ServiceInfo', 'name fullname srvtype rostype_name rostype_req rostype_resp')

Param = namedtuple('Param', 'name value')
# Set :
# rqst : value = <value_asdict>
# resp : value = None
# Get :
# rqst : value = None
# resp : value = <value_asdict>

ParamList = namedtuple('ParamList', 'name_dict')
ParamInfo = namedtuple('ParamInfo', 'name fullname')

Error = namedtuple('Error', 'pid message traceback')





# -*- coding: utf-8 -*-
from __future__ import absolute_import

import warnings
import functools

# This is duplicated from pyros_setup here to not require pyros-setup as dependency when running from deb package.

# http://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  #turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  #reset filter
        return func(*args, **kwargs)

    return new_func
